package com.bazaarvoice.hybris.cronjob;

import com.bazaarvoice.hybris.constants.BazaarvoiceConstants;
import com.bazaarvoice.hybris.exception.BazaarvoiceProductFeedExportException;
import com.bazaarvoice.hybris.model.BazaarvoiceConfigModel;
import com.bazaarvoice.hybris.model.BazaarvoiceTransactionsExportCronJobModel;
import com.bazaarvoice.hybris.utils.BVExportXmlUtils;
import com.bazaarvoice.hybris.xml.PaymentTransactionsXmlCreator;
import com.bazaarvoice.hybris.xml.tags.InteractionTag;
import com.bazaarvoice.hybris.xml.tags.PaymentTransactions;
import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.Session;
import de.hybris.platform.cms2.model.site.CMSSiteModel;
import de.hybris.platform.core.enums.DeliveryStatus;
import de.hybris.platform.core.model.order.AbstractOrderModel;
import de.hybris.platform.cronjob.enums.CronJobResult;
import de.hybris.platform.cronjob.enums.CronJobStatus;
import de.hybris.platform.ordersplitting.model.ConsignmentModel;
import de.hybris.platform.payment.model.PaymentTransactionModel;
import de.hybris.platform.servicelayer.cronjob.AbstractJobPerformable;
import de.hybris.platform.servicelayer.cronjob.PerformResult;
import de.hybris.platform.servicelayer.model.ModelService;
import de.hybris.platform.servicelayer.search.FlexibleSearchQuery;
import de.hybris.platform.servicelayer.search.FlexibleSearchService;
import de.hybris.platform.servicelayer.search.SearchResult;
import de.hybris.platform.servicelayer.session.SessionService;
import de.hybris.platform.util.Config;
import org.apache.commons.net.ftp.FTP;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.*;

@Component("bazaarvoiceTransactionsExporter")
@Scope("prototype")
public class BazaarvoiceTransactionPerformable extends AbstractJobPerformable<BazaarvoiceTransactionsExportCronJobModel> {

    private static final String XMLNS = "http://www.bazaarvoice.com/xs/PRR/PostPurchaseFeed/14.7";
    private static final String XML_BASE_FILE_NAME = "bv_ppe_tag_feed_";

    private static final int TRANSACTIONS_MAX_NUMBER = 75000;

    private BazaarvoiceConfigModel bazaarvoiceConfig;

    private static final Logger LOG = Logger.getLogger(BazaarvoiceTransactionPerformable.class);

    @Resource
    private PaymentTransactionsXmlCreator paymentTransactionsXmlCreator;

    @Resource(name = "bvExportXmlUtils")
    private BVExportXmlUtils bvExportXmlUtils;

    @Override
    public PerformResult perform(final BazaarvoiceTransactionsExportCronJobModel cronJobModel) {
        CMSSiteModel cmsSiteModel = cronJobModel.getBazaarvoiceStorefront();
        bazaarvoiceConfig = cmsSiteModel.getBvConfig();
        String localXmlOutputDirectory = Config.getParameter(BazaarvoiceConstants.LOCAL_XML_OUTPUT_DIRECTORY);
        String specificXmlOutputDirectory = Config.getParameter(BazaarvoiceConstants.SPECIFIC_XML_OUTPUT_DIRECTORY);
        try {
            List<PaymentTransactionModel> result = getPaymentTransactions(cronJobModel);
            List<InteractionTag> interactionTags = paymentTransactionsXmlCreator.createInteractionTagFromCollection(
                    result, cronJobModel.getBazaarvoiceStorefront());

            final List<File> files = new ArrayList<File>();
            for (int i = 0; i * TRANSACTIONS_MAX_NUMBER < interactionTags.size(); i++) {
                final String dateNumber = bvExportXmlUtils.gerenateDateWithMillisecond();
                String siteNameForFileName = cronJobModel.getBazaarvoiceStorefront().getName();
                siteNameForFileName = siteNameForFileName.replaceAll("\\s+", "");
                final String fileName = XML_BASE_FILE_NAME + siteNameForFileName + "_" + dateNumber;
                final File pathToDir = bvExportXmlUtils.getPathToTransactionDir(localXmlOutputDirectory, specificXmlOutputDirectory);
                final File file = new File(pathToDir.getAbsolutePath(), fileName + ".xml");
                if (!file.exists()) {
                    try {
                        boolean created = file.createNewFile();
                        LOG.warn("File created " + created);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }

                PaymentTransactions paymentTransactions = new PaymentTransactions();
                paymentTransactions.setXmlns(XMLNS);

                final Collection<InteractionTag> partInteractions = new ArrayList<>();
                partInteractions.addAll(interactionTags.subList(TRANSACTIONS_MAX_NUMBER * i,
                        TRANSACTIONS_MAX_NUMBER * (i + 1) <= interactionTags.size() ? TRANSACTIONS_MAX_NUMBER * (i + 1) : interactionTags.size()));
                paymentTransactions.setInteractionTags(partInteractions);
                paymentTransactionsXmlCreator.writeXmlToFile(file, paymentTransactions);
                files.add(file);
            }
            if (!files.isEmpty()) {
                sendToFtp(cronJobModel, files);
            } else {
                LOG.info("There are no transactions to export.");
            }
        } catch (final Exception ex) {
            LOG.error("bazaarvoiceTransactionsExporter finished with status FAILURE");
            ex.printStackTrace();
            return new PerformResult(CronJobResult.FAILURE, CronJobStatus.FINISHED);
        }

        return new PerformResult(CronJobResult.SUCCESS, CronJobStatus.FINISHED);
    }

    private void sendToFtp(BazaarvoiceTransactionsExportCronJobModel cronJobModel, List<File> files) {
        //transfering to bazaarvoice ftp server
        String server = bazaarvoiceConfig.getFtpServer();
        final boolean transferResult;
        if(server.toLowerCase().substring(0,4).equals("sftp")){
            transferResult = transferFileToSFTPServer(files, cronJobModel);
        } else {
            transferResult = transferFileToServer(files, cronJobModel);
        }

        if (transferResult)
        {
            LOG.debug("Zip transferred to server SUCCESSFULLY");
        }
        else
        {
            LOG.debug("Zip transferred to server NOT SUCCESSFULLY");
            throw new BazaarvoiceProductFeedExportException("Transfer failed");
        }
    }

    private List<PaymentTransactionModel> getPaymentTransactions(BazaarvoiceTransactionsExportCronJobModel cronJobModel) {
        if (cronJobModel.getExportYesterdaysTransactions()){
             return getYesterdaysTransactions(cronJobModel);
        } else {
            if (cronJobModel.getFilterByShippingDate()) {
                return getTransactionsByShippingDate(cronJobModel);
            } else {
                return getTransactionsByTransactionDate(cronJobModel);
            }
        }
    }

    private List<PaymentTransactionModel> getTransactionsByShippingDate(BazaarvoiceTransactionsExportCronJobModel cronJobModel) {
        FlexibleSearchQuery allTransactionsFlexQuery = new FlexibleSearchQuery("Select {pk} from {PaymentTransaction as p}");
        SearchResult<PaymentTransactionModel> allTransactions = flexibleSearchService.search(allTransactionsFlexQuery);
        if (cronJobModel.getStartDate() != null && cronJobModel.getEndDate() != null) {
            return getTransactionsByStartAndEndDate(cronJobModel, allTransactions);
        } else if (cronJobModel.getStartDate() != null && cronJobModel.getEndDate() == null) {
            return getTransactionsByStartDate(cronJobModel, allTransactions);
        } else if (cronJobModel.getStartDate() == null && cronJobModel.getEndDate() != null) {
            return getTransactionsByEndDate(cronJobModel, allTransactions);
        } else {
            return allTransactions.getResult();
        }
    }

    private List<PaymentTransactionModel> getTransactionsByEndDate(BazaarvoiceTransactionsExportCronJobModel cronJobModel, SearchResult<PaymentTransactionModel> allTransactions) {
        List<PaymentTransactionModel> result = new ArrayList<>();
        for (PaymentTransactionModel transaction : allTransactions.getResult()) {
            Set<ConsignmentModel> consignments = getConsignments(cronJobModel, result, transaction);
            for (ConsignmentModel consignment : consignments) {
                Date shippingDate = consignment.getShippingDate();
                if (shippingDate != null && (shippingDate.before(cronJobModel.getEndDate()) || shippingDate.equals(cronJobModel.getEndDate()))) {
                    result.add(transaction);
                }
            }
        }
        return result;
    }

    private List<PaymentTransactionModel> getTransactionsByStartDate(BazaarvoiceTransactionsExportCronJobModel cronJobModel, SearchResult<PaymentTransactionModel> allTransactions) {
        List<PaymentTransactionModel> result = new ArrayList<>();
        for (PaymentTransactionModel transaction : allTransactions.getResult()) {
            Set<ConsignmentModel> consignments = getConsignments(cronJobModel, result, transaction);
            for (ConsignmentModel consignment : consignments) {
                Date shippingDate = consignment.getShippingDate();
                if (shippingDate != null && (shippingDate.after(cronJobModel.getStartDate()) || shippingDate.equals(cronJobModel.getStartDate()))) {
                    result.add(transaction);
                }
            }
        }
        return result;
    }

    private List<PaymentTransactionModel> getTransactionsByStartAndEndDate(BazaarvoiceTransactionsExportCronJobModel cronJobModel, SearchResult<PaymentTransactionModel> allTransactions) {
        List<PaymentTransactionModel> result = new ArrayList<>();
        for (PaymentTransactionModel transaction : allTransactions.getResult()) {
            Set<ConsignmentModel> consignments = getConsignments(cronJobModel, result, transaction);
            for (ConsignmentModel consignment : consignments) {
                Date shippingDate = consignment.getShippingDate();
                if (shippingDate != null && (shippingDate.after(cronJobModel.getStartDate()) && shippingDate.before(cronJobModel.getEndDate()))) {
                    result.add(transaction);
                }
            }
        }
        return result;
    }

    private Set<ConsignmentModel> getConsignments(BazaarvoiceTransactionsExportCronJobModel cronJobModel, List<PaymentTransactionModel> result, PaymentTransactionModel transaction) {
        AbstractOrderModel order = transaction.getOrder();
        if (order != null && order.getDeliveryStatus() != null) {
            if (order.getDeliveryStatus().equals(DeliveryStatus.SHIPPED)) {
                Set<ConsignmentModel> consignments = order.getConsignments();
                if (CollectionUtils.isEmpty(consignments)) {
                    LOG.debug("Order with code: [" + order.getCode() + "] and status 'Shipped' does not " +
                            "have consignments, hence can not be included into filtering by shipping date");
                } else {
                    return consignments;
                }
            }
        }
        return new HashSet<>();
    }

    private List<PaymentTransactionModel> getTransactionsByTransactionDate(BazaarvoiceTransactionsExportCronJobModel cronJobModel) {
        FlexibleSearchQuery flexQuery;
        if (cronJobModel.getStartDate() != null && cronJobModel.getEndDate() != null) {
            String queryString = "Select {pk} from {PaymentTransaction as p} WHERE {p.creationtime} between ?startDate and ?endDate";
            final Map<String, Object> queryParameters = new HashMap<>();
            queryParameters.put("startDate", cronJobModel.getStartDate());
            queryParameters.put("endDate", cronJobModel.getEndDate());
            flexQuery = new FlexibleSearchQuery(queryString, queryParameters);
        } else if (cronJobModel.getStartDate() != null && cronJobModel.getEndDate() == null) {
            String queryString = "Select {pk} from {PaymentTransaction as p} WHERE {p.creationtime} >= ?startDate";
            final Map<String, Object> queryParameters = new HashMap<>();
            queryParameters.put("startDate", cronJobModel.getStartDate());
            flexQuery = new FlexibleSearchQuery(queryString, queryParameters);
        } else if (cronJobModel.getStartDate() == null && cronJobModel.getEndDate() != null) {
            String queryString = "Select {pk} from {PaymentTransaction as p} WHERE {p.creationtime} <= ?endDate";
            final Map<String, Object> queryParameters = new HashMap<>();
            queryParameters.put("endDate", cronJobModel.getEndDate());
            flexQuery = new FlexibleSearchQuery(queryString, queryParameters);
        } else {
            String queryString = "Select {pk} from {PaymentTransaction as p}";
            flexQuery = new FlexibleSearchQuery(queryString);
        }
        SearchResult<PaymentTransactionModel> searchResult = flexibleSearchService.search(flexQuery);
        return searchResult.getResult();
    }

    private List<PaymentTransactionModel> getYesterdaysTransactions(BazaarvoiceTransactionsExportCronJobModel cronJobModel) {
        LocalDate today = LocalDate.now();
        Date todayDate = Date.from(today.atStartOfDay(ZoneId.systemDefault()).toInstant());
        LocalDate yesterday = today.minusDays(1);
        Date yesterdayDate = Date.from(yesterday.atStartOfDay(ZoneId.systemDefault()).toInstant());

        String queryString = "Select {pk} from {PaymentTransaction as p} WHERE {p.creationtime} >= ?yesterdayDate AND {p.creationtime} < ?todayDate";
        final Map<String, Object> queryParameters = new HashMap<>();
        queryParameters.put("yesterdayDate", yesterdayDate);
        queryParameters.put("todayDate", todayDate);
        FlexibleSearchQuery flexQuery = new FlexibleSearchQuery(queryString, queryParameters);
        SearchResult<PaymentTransactionModel> searchResult = flexibleSearchService.search(flexQuery);
        return searchResult.getResult();
    }

    private boolean transferFileToServer(final List<File> files, BazaarvoiceTransactionsExportCronJobModel cronJobModel)
    {

        boolean result = true;
        final FTPClient client = new FTPClient();
        FileInputStream fis = null;
        String server = null;
        Integer port = null;
        try
        {
            server = bazaarvoiceConfig.getFtpServer();
            String ftpServerPort = Config.getParameter(BazaarvoiceConstants.FTP_SERVER_PORT);
            port = bvExportXmlUtils.getPort(server, ftpServerPort);
            //			server = paramT.getBazaarvoiceConfig().getFtpServer();//bazaarvoiceSettingService.getProperty(BazaarvoiceConstants.FTP_STAGING_SERVER);
            //			port = paramT.getBazaarvoiceConfig().getFtpTransferPort();
            client.connect(server, port != null ? port.intValue() : 21);
            LOG.debug("Connection established...");
            final boolean login = client.login(bazaarvoiceConfig.getFtpUserName(), bazaarvoiceConfig.getFtpPassword());

            if (login)
            {

                LOG.debug("Client logged...");
                client.enterLocalPassiveMode();

                LOG.debug("Transfering...");
                client.setFileType(FTP.BINARY_FILE_TYPE);
                for (final File file : files)
                {
                    fis = new FileInputStream(file.getAbsoluteFile());
                    LOG.info("Trying to transfer file " + file.getAbsolutePath());
                    final String serverFileName = cronJobModel.getFtpServerDirectory() + "/" + file.getName();
                    LOG.debug("File name on server: " + serverFileName);
                    final boolean done = client.storeFile(serverFileName, fis);
                    result &= done;
                }

                final boolean logout = client.logout();
                if (logout)
                {
                    LOG.debug("Connection close...");
                }
            }
            else
            {
                result = false;
                LOG.debug("Logging failed...");
            }

        }
        catch (final Exception e)
        {
            result = false;
            LOG.debug("Connection failed...");
            if (server == null || port == null)
            {
                LOG.debug("Please check your Bazaarvoice settings! If they are absent, update your system or import them!");
            }
            e.printStackTrace();
            LOG.debug("Transfer failed");
        }
        finally
        {
            try
            {
                fis.close();
            }
            catch (final IOException e)
            {
                e.printStackTrace();
            }
            try
            {
                client.disconnect();
            }
            catch (final IOException e)
            {
                e.printStackTrace();
            }

        }
        return result;
    }

    private boolean transferFileToSFTPServer(final List<File> files, BazaarvoiceTransactionsExportCronJobModel cronJobModel)
    {
        Session session = null;
        Channel channel = null;
        ChannelSftp channelSftp = null;
        Integer port = null;
        String server = null;
        LOG.info("preparing the host information for sftp.");
        try {
            server = bazaarvoiceConfig.getFtpServer();
            String ftpServerPort = Config.getParameter(BazaarvoiceConstants.FTP_SERVER_PORT);
            port = bvExportXmlUtils.getPort(server, ftpServerPort);
            JSch jsch = new JSch();
            session = jsch.getSession(bazaarvoiceConfig.getFtpUserName(), bazaarvoiceConfig.getFtpServer(), port);
            session.setPassword(bazaarvoiceConfig.getFtpPassword());
            java.util.Properties config = new java.util.Properties();
            config.put("StrictHostKeyChecking", "no");
            session.setConfig(config);
            session.connect();
            LOG.info("Host connected.");
            channel = session.openChannel("sftp");
            channel.connect();
            LOG.info("sftp channel opened and connected.");
            channelSftp = (ChannelSftp) channel;
            for (final File file : files)
            {
                LOG.info("Trying to transfer file " + file.getAbsolutePath());
                channelSftp.cd(cronJobModel.getFtpServerDirectory());
                channelSftp.put(new FileInputStream(file), file.getName());
                LOG.info("File name on server: " + file.getName());
            }

        } catch (Exception ex) {
            LOG.info("Exception found while tranfer the response.");
            ex.printStackTrace();
            return false;
        }
        finally{

            channelSftp.exit();
            LOG.info("sftp Channel exited.");
            channel.disconnect();
            LOG.info("Channel disconnected.");
            session.disconnect();
            LOG.info("Host Session disconnected.");
        }
        return true;
    }

    @Override
    @Autowired(required = true)
    public void setModelService(final ModelService modelService) {

        this.modelService = modelService;
    }

    @Override
    @Autowired(required = true)
    public void setFlexibleSearchService(final FlexibleSearchService flexibleSearchService) {

        this.flexibleSearchService = flexibleSearchService;
    }

    @Override
    @Autowired(required = true)
    public void setSessionService(final SessionService sessionService) {

        this.sessionService = sessionService;
    }
}
